import socket
import time
import argparse
import sys
import json

maxWaitTime = 3

class CnfmonClient:

    debug = False
    verbose = False
    server_address = '/run/cnfmon.sock'

    def __init__(self, args):
        if 'socket' in args:
            args['server_address'] = args['socket']
        for name in args:
            if hasattr(self, name) and type(getattr(self, name)) in [str,float,int,bool]:
                if self.debug:
                    print('Setting property: %s=%s' % (name, args[name]))
                setattr(self, name, args[name])
        self.connect()

    def connect(self):

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        t0 = time.time()

        while True:
            if self.debug:
                print('connecting to {}'.format(self.server_address))
            try:
                sock.connect(self.server_address)
                break
            except socket.error as msg:
                failmsg = msg
            if time.time() - t0 > maxWaitTime:
                print('Failed to connect to server in %g s' % (maxWaitTime,))
                print(failmsg)
                sys.exit(1)
            time.sleep(0.5)

        if self.debug:
            print('connected to {}'.format(self.server_address))

        sock.settimeout(0.1)
        self.sockfd = sock

    def sendcmd(self, cmd, **kw):
        kw['cmd'] = cmd
        cd = json.dumps(kw)
        msg = '%s\n' % (cd,)
        msg = msg.encode('utf8')
        self.sockfd.send(msg)
        chunk = resp = b''
        while True:
            try:
                chunk = self.sockfd.recv(1024)
            except:
                break
                pass
            if len(chunk) == 0 or chunk == None:
                break
            resp += chunk
        resp = resp.decode('utf8')
        if self.verbose:
            print(resp)
        return resp

    def execute(self, path):
        self.sendcmd('run', path=path)

    def status(self):
        self.sendcmd('status')

    def scan(self):
        self.sendcmd('scan')

    def update(self):
        self.sendcmd('update')

    def clear(self):
        self.sendcmd('clear')

    def reset(self):
        self.sendcmd('reset')

    def quitServer(self):
        self.sendcmd('quit')

    def close(self):
        self.sendcmd('bye')

    def version(self):
        self.sendcmd('version')

    def interactive(self):
        while not sys.stdout.closed:
            sys.stdout.write('gcli> ')
            sys.stdout.flush()
            ln = sys.stdin.readline()
            if len(ln) == 0:
                print('input closed, exit')
                break
            if len(ln.strip()) == 0:
                print('empty line, ignore')
                continue
            print('cmd: "%s"' % (ln,))
            self.sockfd.write('%s' % (ln,))
            self.sockfd.flush()
            resp = self.sockfd.readline()
            print(resp)

def parseargs():
    import argparse
    import configparser

    conf_parser = argparse.ArgumentParser(add_help=False)
    conf_parser.add_argument("-c", "--conf_file", default=["/etc/cnfmon.conf", "cnfmon.conf",
                                                           "/etc/cnfmonclient.conf", "cnfmonclient.conf"],
                             help="Specify config file", metavar="FILE")
    conf_parser.add_argument("-C", "--conf_section", default="cnfmonclient",
                             help="Specify config file section", metavar="NAME")
    args, remaining_argv = conf_parser.parse_known_args()

    defaults = {
        'debug': False,
        'socket': '/run/cnfmon.sock',
        'logfile': 'cnfmonclient.log',
    }
    if args.conf_file:
        config = configparser.ConfigParser(defaults=defaults)
        if False:
            print('defaults', defaults)
            print("Reading configuration file from %s" % (args.conf_file,))
        cread = config.read(args.conf_file)
        if False:
            print("Configuration files read: %s" % (cread,))
        if len(cread):
            if args.conf_section in config:
                defaults = dict(config.items(args.conf_section))
                if False:
                    print('Loading configuration defaults from %s/[%s]' % (cread, args.conf_section))
                    print('defaults', defaults)

    parser = argparse.ArgumentParser(description='GPIO server client',
                                     parents=[conf_parser],
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--debug', action='store_true', help='Debug')
    parser.add_argument('--socket', type=str, help='Server socket name')

    parser.add_argument('--clear', action='store_true', help='Clear')
    parser.add_argument('--reset', action='store_true', help='Reset')
    parser.add_argument('--status', action='store_true', help='Status')
    parser.add_argument('--server-version', action='store_true', help='Server version')
    parser.add_argument('--scan', action='store_true', help='Scan now')
    parser.add_argument('--update', action='store_true', help='Update')

#    parser.add_argument('path', type=str, metavar='prog', nargs='?', help='Program to execute by cnfmon server')

    parser.set_defaults(**defaults)

    args = parser.parse_args(remaining_argv)

    if args.debug:
        print('parsed args', vars(args))
    return args

def run():
    args = parseargs()
    cli = CnfmonClient(vars(args))
    cli.verbose = True

    if args.status:
        cli.status()
    elif args.clear:
        cli.clear()
    elif args.reset:
        cli.reset()
    elif args.update:
        cli.update()
    elif args.scan:
        cli.scan()
    elif args.server_version:
        cli.version()
    elif args.path:
        cli.execute(args.path)
    else:
        print('nothing to do')

if __name__ == "__main__":
    run()
