import os
import time
import signal
import stat
import socket
import threading
import json
import time
import subprocess
import sys
import shutil
import configparser

from . import VERSION

class CnfmonServer:
    conf_file = ''
    server_address = '/run/cnfmon.sock'
    connections = {}
    debug = False
    verbose = False
    stop = False
    pairs = []
    monpairs = []

    def __init__(self, args={}):
        if 'socket' in args:
            args['server_address'] = args['socket']
        for name in args:
            if hasattr(self, name) and type(getattr(self, name)) in [str,float,int,bool,list]:
                print('Setting property: %s=%s' % (name, args[name]))
                setattr(self, name, args[name])

    def openSocket(self):
        sockdir = os.path.dirname(self.server_address)
        try:
            os.mkdir(sockdir)
        except OSError:
            if not os.path.exists(sockdir):
                raise

        # Make sure the socket does not already exist
        try:
            os.unlink(self.server_address)
        except OSError:
            if os.path.exists(self.server_address):
                raise

        # Create a UDS socket
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        # Bind the socket to the address
        print('bind to socket %s' % (self.server_address,))
        sock.bind(self.server_address)

        os.chmod(self.server_address, stat.S_IRWXU|stat.S_IRWXG|stat.S_IRWXO)

        # Listen for incoming connections
        sock.listen(10)

        # Set timeout
        sock.settimeout(0.5)

        self.sock = sock

    def readpairs(self):
        config = configparser.ConfigParser()
        print("Update: Reading configuration file from %s" % (self.conf_file[0],))
        cread = config.read(self.conf_file[0])
        self.monpairs = []
        for p in self.pairs:
            pdef = dict(config.items(p))
            print(pdef)
            if 'src' in pdef and 'dest' in pdef:
                self.monpairs += [pdef]
        print('updated conf:', self.monpairs)

    def isNewer(self, f1, f2):
        s1 = None
        s2 = None

        try:
            s1 = os.stat(f1)
        except BaseException as E:
            print('Failed to stat file: %s', (E,))
        try:
            s2 = os.stat(f2)
        except BaseException as E:
            print('Failed to stat file: %s', (E,))

        if s2 is not None and s2.st_size > 0:
            if s1 is None or s2.st_mtime > s1.st_mtime:
                return True

        return False

    def scan(self):
        if self.verbose:
            print('scanning pairs...', self.monpairs)
        for p in self.monpairs:
            f1, f2 = p['dest'], p['src']
            if self.verbose:
                print('Checking: "%s" vs. "%s"' % (f2, f1))
            if self.isNewer(f1, f2):
                print('Copy configuration file "%s" -> "%s"' % (f2, f1))
                shutil.copyfile(f2, f1)
                os.chmod(f1, stat.S_IRUSR | stat.S_IWUSR)
        if self.verbose:
            print('scan complete', self.monpairs)

    def clear(self):
        pass

    def reset(self):
        pass

    def handleClosed(self):
        ckeys = list(self.connections.keys())
        for c in ckeys:
            entry = self.connections[c]
            if entry['conn'].closed:
                entry['thread'].join()
                print('Connection id %d is closed, thread joined' % (c,))
                del self.connections[c]
        if self.debug:
            print('Open connections: %s' % (self.connections,))

    def server(self):

        cid = 1

        while not self.stop:
            try:
                connection, client_address = self.sock.accept()
            except socket.timeout as ex:
                self.handleClosed()
                self.scan()
                continue
            except BaseException as ex:
                print(ex)
                raise

            connection.setblocking(True)
            confile = connection.makefile('rw')

            print('new connection from %s: %s, %s' %(client_address, connection.getblocking(), connection.gettimeout()))

            thread = threading.Thread(target=self.handle, args=(confile, client_address))
            self.connections[cid] = {'thread': thread, 'conn': confile}
            cid += 1
            thread.start()

            # self.handle(connection.makefile('rw'), client_address)


    def handle(self, conn, client_address):
        while not conn.closed:
            line = conn.readline().strip()
            if len(line) == 0:
                print('connection closed by %s' % (client_address,))
                try:
                    conn.close()
                except:
                    pass
                break

            try:
                args = json.loads(line)
            except BaseException as e:
                print(e)
                conn.write('invalid cmmand data: %s\n' % (e,))
                conn.flush()
                continue

            if self.verbose:
                print('cmd %s from connection %s' % (args, client_address,))

            if 'cmd' not in args:
                print('no cmd in data: %s' %(cmd,))
            cmd = args['cmd']

            if cmd == 'status':
                r = 'Cnfmon server running\n'
                r += f'Pair names:{self.pairs}\n'
                r += f'File pairs:{self.monpairs}'
                conn.write(r)

            elif cmd == 'scan':
                self.scan()
                conn.write('scan\n')

            elif cmd == 'update':
                self.readpairs()
                conn.write('update\n')

            elif cmd == 'clear':
                self.clear()
                conn.write('clear\n')

            elif cmd == 'version':
                conn.write('server version: %s\n' % VERSION)

            elif cmd == 'bye' or cmd == 'goodbye':
                conn.write('goodbye\n')
                conn.close()

            elif cmd == 'quit':
                conn.write('server shutting down\n')
                self.stop = True
                conn.close()

            else:
                conn.write('error: unknown command: %s\n' %(line,))

            sys.stdout.flush()
            if not conn.closed:
                try:
                    conn.close()
                except BaseException as e:
                    print(e)


def parseargs():
  import argparse

  conf_parser = argparse.ArgumentParser(add_help=False)
  conf_parser.add_argument("-c", "--conf_file", default=["/etc/cnfmon.conf", "cnfmon.conf"],
                           help="Specify config file", metavar="FILE")
  conf_parser.add_argument("-C", "--conf_section", default="cnfmonserver",
                           help="Specify config file section", metavar="NAME")
  args, remaining_argv = conf_parser.parse_known_args()

  defaults = {
    'debug': 0,
    'verbose': 0,
    'socket': '/run/cnfmon.sock',
    'logfile': 'cnfmon.log',
    'pairs': [],
  }
  if args.conf_file:
    print('defaults', defaults)
    config = configparser.ConfigParser(defaults=defaults)
    print("Reading configuration file from %s" % (args.conf_file,))
    cread = config.read(args.conf_file)
    print("Read configuration files: %s" % (cread,))
    if len(cread):
      defaults = dict(config.items(args.conf_section))
      defaults['pairs'] = defaults['pairs'].split(',')
      print('Loading configuration defaults from %s/[%s]' % (cread, args.conf_section))
      print('defaults', defaults)

  parser = argparse.ArgumentParser(description='GPIO server daemon',
                                   parents=[conf_parser],
                                   formatter_class=argparse.RawDescriptionHelpFormatter)

  parser.add_argument('--debug', type=int, help='Debug')
  parser.add_argument('--verbose', type=int, help='Debug')
  parser.add_argument('--socket', type=str, help='Server socket name')
  parser.add_argument('--logfile', '-o', type=str, help='Log file name')
  parser.add_argument('--bindir', '-b', type=str, help='Executables directory name')
  parser.add_argument('--pairs', '-p', nargs='+', type=str, help='Names pairs of files to monitor')

  parser.set_defaults(**defaults)

  args = parser.parse_args(remaining_argv)

  print('parsed args', vars(args))
  return args

def signalHandler(sig, frame, server):
    print(f'Signal {sig} caught')
    server.stop = True

def run():
    args = parseargs()
    server = CnfmonServer(vars(args))
    server.openSocket()
    signal.signal(signal.SIGINT, lambda a,b: signalHandler(a, b, server))
    server.readpairs()
    server.server()
    print('cnfmon server exiting')

if __name__ == "__main__":
    run()
